#ifndef __DIJKSTRA_CUDA__
#define __DIJKSTRA_CUDA__

#include <stdlib.h>
#include <limits.h>

#define ID 13517122

/**
 * Get vertex index with minimum distance which not yet included
 * in spt_set
 * @param  dist    distance from origin vertex to vertex with that index
 * @param  spt_set a set denoting vertices included in spt_set
 * @param n number of vertices in the graph
 * @return         index of minimum distance not yet included in spt_set
 *//*
long min_distance_idx(long *dist, bool *spt_set, int n) {
	// Initialize min value 
    long min = LONG_MAX, min_index; 
  
    for (int i = 0; i < n; i++) {
    	if (spt_set[i] == false && dist[i] <= min) {
    		min = dist[i];
    		min_index = i;
    	}
    } 
	spt_set[min_index] = true;
	dist[src] = 0;
	
    return min_index; 
}
*/

// for dijkstra algorithm
__global__ 
void initValue(long *graph, long *allResult, int *visitedNode, int *minIndex, int sourceIdx, int num_vertices) {
	int index = threadIdx.x + blockDim.x * blockIdx.x;
	int stride = blockDim.x * gridDim.x;

	for (int i=index; i < num_vertices; i += stride) {
		visitedNode[i] = 0;
		if ((graph[i*num_vertices + sourceIdx]==0) && i!=(sourceIdx)) {
			allResult[i] = LONG_MAX;
		}
		else {
			allResult[i] = graph[i*num_vertices + sourceIdx];
		}
	}
	
	*minIndex = -1;
	visitedNode[sourceIdx] = 1;
}
/*
__global__ 
void findMinDistance(long *allResult, int *visitedNode, int *minIndex, long *minDistance, int num_vertices) {
	
	*minDistance = LONG_MAX;

	for (int j=0; j<num_vertices; j++) {
		if (visitedNode[j]==0 && allResult[j]<*minDistance) {
			*minDistance = allResult[j];
			*minIndex = j;
		}
	}

	visitedNode[*minIndex] = 1;
	
}
*/
__global__
void findAndSetNewDistance(long *graph, long *allResult, int *visitedNode, int *minIndex, long *minDistance, int num_vertices) {
	
	*minDistance = LONG_MAX;

	for (int j=0; j<num_vertices; j++) {
		if (visitedNode[j]==0 && allResult[j]<*minDistance) {
			*minDistance = allResult[j];
			*minIndex = j;
		}
	}
	
	__syncthreads();
	
	visitedNode[*minIndex] = 1;
	
	int index = threadIdx.x + blockDim.x * blockIdx.x;
	int stride = blockDim.x * gridDim.x;

	for (int i=index; i < num_vertices; i += stride) {
		if (visitedNode[i]) {
			continue;
		}
		else if ((graph[i*num_vertices + *minIndex]+*minDistance<allResult[i]) && (graph[i*num_vertices + *minIndex]+*minDistance!=0)) {
			allResult[i] = graph[i*num_vertices + *minIndex]+*minDistance;
		}
	}
}


/**
 * generate a graph with n vertices
 * @param  n number of vertices
 * @return   1D array, graph[i*n + j] = graph[j*n + i] = distance from vertex i to j
 */
 long* create_graph(int n) {
    int i,j;
	long *graph = (long*) malloc(n * n * sizeof(long));

    for (i=0;i<n;i++) {
            for (j=i;j<n;j++) {
                if (i==j) {
                    graph[i*n + j] = 0;
                }
                else {
                    graph[i*n + j] = rand();
                    graph[j*n + i] = graph[i*n + j];
                }
            }
        }
    return graph;
}
//-----
long* create_temp(int n) {
    int i,j;
	long *graph = (long*) malloc(n * n * sizeof(long));
	
    for (i=0;i<n;i++) {
            for (j=i;j<n;j++) {
				graph[i*n + j] = INT_MAX;
            }
        }
    return graph;
}
/*
 long **gen_graph(int n) {
	// alokasi memori untuk matriks yang merepresentasikan graf
	long **result = (long **)malloc(n * sizeof(long *)); 
    for (int i = 0; i < n; i++) {
    	result[i] = (long *)malloc(n * sizeof(long)); 
    }    
  
    // isi matriks dengan bilangan random
    
    
    for (int i = 0; i < n; i++) {
    	for (int j = i; j < n; j++) {
    		if (i == j) {
    			result[i][j] = 0;
    		} else {
    			result[i][j] = result[j][i] = rand();	
    		}
    		
    	}
    }
     
 	return result; 
}*/
//make the graph as graph[i*n + j] to make it able to be malloc on cuda as 1d array
//-----
//-----
/*
long **gen_temp(int r, int c) {
	// alokasi memori untuk matriks yang merepresentasikan graf
	long **result = (long **)malloc(r * sizeof(long *)); 
	// printf("[gen_temp] initiate temp\n");
    for (int i = 0; i < r; i++) {
    	result[i] = (long *)malloc(c * sizeof(long)); 
    	// printf("[gen_temp] initiate each row in temp\n");
    }    
    
    for (int i = 0; i < r; i++) {
    	for (int j = 0; j < c; j++) {
    		// printf("[gen_temp] filling temp\n");
    		result[i][j] = 0;
    	}
    }
     
 	return result; 
}*/
/*
long *dijkstra(long **graph, int n, int src) {

	// output array, contains shortest distance from src to every vertices
	long *dist = (long *) malloc (sizeof(long) * n);
	// spt_set[i] is true if vertex i already included in the shortest path tree
	bool *spt_set = (bool *) malloc(sizeof(bool) * n);

	// initialize dist and spt_set
	for (int i = 0; i < n; i++) {
		dist[i] = INT_MAX;
		spt_set[i] = false;
	}

	// initiate path searching 
	dist[src] = 0;


	// find the shortest path for all vertices
	for (int i = 0; i < n; i++) {

		// pick vertex with minimum distance from src from spt_set not yet
		// processed
		int processed_vertex = min_distance_idx(dist, spt_set, n);

		// mark vertex as processed
		spt_set[processed_vertex] = true;

		for (int j = 0; j < n; j++) {
			// check vertices connected to processed_vertex not yet processed
			if (!spt_set[j] 
				&& graph[processed_vertex][j] != 0
				&& dist[processed_vertex] != INT_MAX 
				&& dist[processed_vertex] + graph[processed_vertex][j] < dist[j]) {

				dist[j] = dist[processed_vertex] + graph[processed_vertex][j];
			}
		}
	}
	free(spt_set);

	return dist;
}
*/
/**
* that one kernel that do it "all"
*/
/*
__global__ void do_it(**long graph, **long result, int num_vertices) {
	int start_idx = threadIdx.x + blockDim.x * blockIdx.x;
	int stride = blockDim.x * gridDim.x;
	
	long *temp = (long*) malloc(sizeof(long)*num_vertices);
	
	for (int i = start_idx; i < num_vertices; i+= stride) {
		temp = dijkstra(graph, num_vertices, i);
		
		for (int j = 0; j < num_vertices; j++) {
			result[i][j] = temp[j];
		}
	}
}
*/

void print_graph(long *data, int n) {
    int i,j;
    for (i=0;i<n;i++) {
        for (j=0;j<n;j++) {
            printf("%li ",data[i*n + j]);
        }
        printf("\n");
    }
}

void write_to_txt(int n, long *const graph, const char* filename) {
    FILE *fout;
    int i,j;
    if (NULL == (fout = fopen(filename,"w"))) {
        fprintf(stderr,"error opening output file");
        abort();
    }

    for (i=0;i<n;i++) {
        for(j=0;j<n;j++) {
            fprintf(fout,"%li ",graph[i*n + j]);
        }
        fprintf(fout,"\n");
    }
    printf("Result has been written to %s ...\n",filename);
}

#endif