#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include "dijkstra_cuda.cuh"

#define THREADS_BLOCK 256

static double get_micros(void) {
    struct timespec ts;
    timespec_get(&ts, TIME_UTC);
    return ((double)((long)ts.tv_sec * 1000000000L + ts.tv_nsec)/1000);
}

int main (int argc, char const *argv[]) {
    
    //check if argc == 3
    if (argc!=3) {
        fprintf(stderr,"Usage: Dijkstra_CUDA num_of_node output_filename\n");
        return EXIT_FAILURE;
    }
    // initialization
    srand(ID);
    int num_vertices = atoi(argv[1]);
    double start_time, end_time, total_time;
    total_time = 0;

    // allocate memory in host for the graph
    long *graph = create_graph(num_vertices);
    

    // allocate memory in the host for the result matrice
    long *result = create_temp(num_vertices);
    
    // allocate memory in the host for result array from a vertice
    long *tempResult = (long *)malloc(num_vertices * sizeof(long));

    for (int i=0; i<num_vertices; i++) {
        tempResult[i] = -1;
    }

    //CUDA malloc initialize
    long *gpu_graph;
    long *gpu_result;
    int *gpu_visitedNode;
    long *minDistance;
    int *minIndex;

    //CUDA malloc
    // allocate memory in device for the graph
    cudaMalloc((void**)&gpu_graph,num_vertices*num_vertices*sizeof(long));
    // allocate memory in device for the result of dijkstra
    cudaMalloc((void**)&gpu_result,num_vertices*sizeof(long));
    // allocate memory in device for the list of visited node
    cudaMalloc((void**)&gpu_visitedNode,num_vertices*sizeof(int));
    // allocate memory in device for the minimal distance used in dijkstra
    cudaMalloc((void**)&minDistance,sizeof(long));
    // allocate memory in device for the index of minDistance
    cudaMalloc((void**)&minIndex,sizeof(int));

    // copy data of graph from host to device
    cudaMemcpy(gpu_graph,graph,num_vertices*num_vertices*sizeof(long),cudaMemcpyHostToDevice);
    
    // initiate block size and num of blocks that will be use in device
    int blockSize = 256;
    int numBlocks = (num_vertices + blockSize - 1) / blockSize;

    // dijkstra algorithm for each vertice
    for (int i=0; i<num_vertices; i++) {
        
        // initialize value for dijkstra in device
        initValue<<<numBlocks, blockSize>>>(
            gpu_graph,
            gpu_result,
            gpu_visitedNode,
            minIndex,
            i,
            num_vertices);
            
        // set timer
        start_time = get_micros();
        // for each vertice except current vertice (source)
        
        for (int j=1; j<num_vertices; j++) {
            
            // find minimal distance
            //findMinDistance<<<numBlocks,blockSize,num_vertices*sizeof(int)+num_vertices*sizeof(long)>>>(
            //
            //findMinDistance<<<1,1>>>(
            //    gpu_result,
            //    gpu_visitedNode,
            //    minIndex,
            //    minDistance,
            //    num_vertices);
            //
            // update distance for each vertice if new distance < old distance
            findAndSetNewDistance<<<numBlocks, blockSize>>>(
                gpu_graph,
                gpu_result,
                gpu_visitedNode,
                minIndex,
                minDistance,
                num_vertices);
            
            // cudaDeviceSynchronize();
        }
        
        // end of timer
        end_time = get_micros();

        // copy the result from device to host
        cudaMemcpy(tempResult,gpu_result,num_vertices*sizeof(long),cudaMemcpyDeviceToHost);

        // fill copied into the result matrice
        for (int k=0; k<num_vertices; k++) {
            result[i*num_vertices + k] = tempResult[k];
        }

        total_time += end_time-start_time;
    }
    
    
    write_to_txt(num_vertices,result,argv[2]);
    
    printf("processing time: %0.04lf us ...\n",total_time);

    // free device memory allocation
    cudaFree(gpu_graph);
    cudaFree(gpu_result);
    cudaFree(gpu_visitedNode);
    cudaFree(minDistance);
    cudaFree(minIndex);

    // free host memory allocation
    free(tempResult);
    free(graph);
    free(result);

    return EXIT_SUCCESS;
}